/* case.cc - Convert case in strings
 * Copyright 2008 Bas Wijnen <wijnen@debian.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "case.hh"
#include <ctype.h>

namespace shevek
{
	std::string toupper (std::string const &orig)
	{
		std::string ret;
		for (std::string::const_iterator
				i = orig.begin (); i != orig.end (); ++i)
			ret += ::toupper (*i);
		return ret;
	}

	std::string tolower (std::string const &orig)
	{
		std::string ret;
		for (std::string::const_iterator
				i = orig.begin (); i != orig.end (); ++i)
			ret += ::tolower (*i);
		return ret;
	}
}
